#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

import template

_DIRECTION_INPUT = True
_DIRECTION_OUTPUT = False


class IOMode(object):

    """
    >>> mode = IOMode()
    >>> mode.is_input()
    True
    >>> mode.get_prompt() == template.getinputprompt()
    True
    >>> mode.set_output()
    >>> mode.is_output()
    True
    >>> mode.get_prompt() == template.getoutputprompt()
    True
    """
    _direction = _DIRECTION_INPUT

    def is_input(self):
        return self._direction == _DIRECTION_INPUT

    def is_output(self):
        return self._direction == _DIRECTION_OUTPUT

    def set_input(self):
        self._direction = _DIRECTION_INPUT

    def set_output(self):
        self._direction = _DIRECTION_OUTPUT

    def get_prompt(self):
        if self.is_input():
            return template.getinputprompt()
        return template.getoutputprompt()


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



#!/usr/bin/env python
# -*- coding:utf-8
import unittest
import doctest


class DocTest(unittest.TestCase):

    def test_tff(self):
        import trachet.tffstub as tffstub
        from trachet.tffstub import tff
        print tffstub.expected_hash
        print tff.signature
        print tff.__version__
        #self.assertTrue(tffstub.expected_hash == tff.signature)

    def test_templete(self):
        import trachet.template as template
        template.enable_color()
        failure_count, test_count = doctest.testmod(template)
        self.assertTrue(failure_count == 0)

    def test_seqdb(self):
        import trachet.template as template
        import trachet.seqdb as seqdb
        template.enable_color()
        failure_count, test_count = doctest.testmod(seqdb)
        self.assertTrue(failure_count == 0)

    def test_iomode(self):
        import trachet.template as template
        import trachet.iomode as iomode
        template.enable_color()
        failure_count, test_count = doctest.testmod(iomode)
        self.assertTrue(failure_count == 0)

    def test_char(self):
        import trachet.template as template
        import trachet.char as char
        template.enable_color()
        failure_count, test_count = doctest.testmod(char)
        self.assertTrue(failure_count == 0)

    def test_esc(self):
        import trachet.template as template
        import trachet.esc as esc
        template.enable_color()
        failure_count, test_count = doctest.testmod(esc)
        self.assertTrue(failure_count == 0)

    def test_csi(self):
        import trachet.template as template
        import trachet.csi as csi
        template.enable_color()
        failure_count, test_count = doctest.testmod(csi)
        self.assertTrue(failure_count == 0)

    def test_cstr(self):
        import trachet.template as template
        import trachet.cstr as cstr
        template.enable_color()
        failure_count, test_count = doctest.testmod(cstr)
        self.assertTrue(failure_count == 0)

    def test_input(self):
        import trachet.template as template
        import trachet.input as input
        template.enable_color()
        failure_count, test_count = doctest.testmod(input)
        self.assertTrue(failure_count == 0)

    def test_output(self):
        import trachet.template as template
        import trachet.output as output
        template.enable_color()
        failure_count, test_count = doctest.testmod(output)
        self.assertTrue(failure_count == 0)

    def test_trace(self):
        import trachet.template as template
        import trachet.trace as trace
        template.enable_color()
        failure_count, test_count = doctest.testmod(trace)
        self.assertTrue(failure_count == 0)

    def test_controller(self):
        import trachet.template as template
        import trachet.controller as controller
        template.enable_color()
        failure_count, test_count = doctest.testmod(controller)
        self.assertTrue(failure_count == 0)

    def test_ss2(self):
        import trachet.template as template
        import trachet.ss2 as ss2
        template.enable_color()
        failure_count, test_count = doctest.testmod(ss2)
        self.assertTrue(failure_count == 0)

    def test_ss3(self):
        import trachet.template as template
        import trachet.ss3 as ss3
        template.enable_color()
        failure_count, test_count = doctest.testmod(ss3)
        self.assertTrue(failure_count == 0)

    def test_constant(self):
        import trachet.template as template
        import trachet.constant as constant
        template.enable_color()
        failure_count, test_count = doctest.testmod(constant)
        self.assertTrue(failure_count == 0)

    def suite():
        suite = unittest.TestSuite()
        suite.addTests(unittest.makeSuite(DocTest))
        return suite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****


from tffstub import tff


class InputHandler(tff.DefaultHandler):

    ''' <F6> toggle trace state
        <F7> toggle break state
        <F8> step to next char or seq
        <F9> step to next ESC or CSI seq
    '''

    def __init__(self, controller, tracer):
        self._controller = controller
        self._tracer = tracer 

    def handle_esc(self, context, intermediate, final):
        self._tracer.set_input()
        return self._tracer.handle_esc(context, intermediate, final)

    def handle_csi(self, context, parameter, intermediate, final):
        self._tracer.set_input()
        return self._tracer.handle_csi(context, parameter, intermediate, final)

    def handle_ss2(self, context, final):
        self._tracer.set_input()
        return self._tracer.handle_ss2(context, final)

    def handle_ss3(self, context, final):
        self._tracer.set_input()
        return self._tracer.handle_ss3(context, final)

    def handle_control_string(self, context, prefix, value):
        self._tracer.set_input()
        return self._tracer.handle_control_string(context, prefix, value)

    def handle_char(self, context, final):
        self._tracer.set_input()
        return self._tracer.handle_char(context, final)

    def handle_invalid(self, context, seq):
        self._tracer.set_input()
        return self._tracer.handle_invalid(context, seq)

    def handle_draw(self, context):
        self._controller.tick()


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

import seqdb
import template

_DB = seqdb.get()


def get_mnemonic(p, v, is_input):
    '''
      >>> _create_mock_db()
      >>> get_mnemonic(']', '0', False)
      'OSC 0 / set icon name and window title'
      >>> get_mnemonic(']', '4', False)
      'OSC 4 / get or set color palette'
    '''

    if is_input:
        direction = '<'
    else:
        direction = '>'

    key = "%s ESC %s%s<ST>" % (direction, p, v)
    if key in _DB:
        return _DB[key]

    if v[0].isalpha():
        key = "%s ESC %s%s<ST>" % (direction, p, v[0])
        if key in _DB:
            return _DB[key]
    else:

        params = v.split(";")
        length = len(params)

        if length > 3:
            key = "%s ESC %s%s;%s;%s;%s<ST>" % (direction, p,
                                                params[0],
                                                params[1],
                                                params[2],
                                                params[3])
            if key in _DB:
                return _DB[key]
        if length > 2:
            key = "%s ESC %s%s;%s;%s<ST>" % (direction, p,
                                             params[0],
                                             params[1],
                                             params[2])
            if key in _DB:
                return _DB[key]
        if length > 1:
            key = "%s ESC %s%s;%s<ST>" % (direction, p,
                                          params[0],
                                          params[1])
            if key in _DB:
                return _DB[key]
        if length > 0:
            key = "%s ESC %s%s<ST>" % (direction, p, params[0])
            if key in _DB:
                return _DB[key]

        import re
        pattern = "^([\x30-\x3f]*)([\x20-\x30]*[\x40-\x7e])([0-9A-Fa-f]*)"
        pbytes, ifbytes, vbytes = re.search(pattern, v).groups()
        if ifbytes:
            if vbytes:
                key = "%s ESC %s%s%s%s<ST>" % (direction, p, pbytes, ifbytes, vbytes)
                if key in _DB:
                    return _DB[key]

            if pbytes:
                key = "%s ESC %s%s%s<ST>" % (direction, p, pbytes, ifbytes)
                if key in _DB:
                    return _DB[key]

            key = "%s ESC %s%s<ST>" % (direction, p, ifbytes)
            if key in _DB:
                return _DB[key]

    key = "%s ESC %s<ST>" % (direction, p)
    if key in _DB:
        return _DB[key]

    return '[ESC ' + p + ']'


def format_seq(prefix, value, is_input, tracer, controller):
    """
      >>> _create_mock_db()
      >>> template.enable_color()
      >>> format_seq(ord("]"), map(ord, "abcde"), False, None, None).replace("\x1b", "\\x1b")
      u'\\x1b[0;1;37;44m ESC ] \\x1b[0;1;35mabcde \\x1b[37;44mST\\x1b[0;1;36m  OSC / operating system command\\x1b[m'

      >>> format_seq(ord("Q"), map(ord, "cdefg"), False, None, None)
      u'\\x1b[0;1;37;44m ESC Q \\x1b[0;1;35mcdefg \\x1b[37;44mST\\x1b[0;1;36m  [ESC Q]\\x1b[m'
    """

    try:
        v = u''.join(map(unichr, value))
    except OverflowError:
        v = str(value)
    p = chr(prefix)

    mnemonic = get_mnemonic(p, v, is_input)
    if mnemonic[0] == "!":
        return eval(mnemonic[1:])
    return template.getcstr() % (p, v, mnemonic)


def _create_mock_db():
    global _DB
    _DB = {
        '> ESC ]<ST>': 'OSC / operating system command',
        '> ESC ]0<ST>': 'OSC 0 / set icon name and window title',
        '> ESC ]1<ST>': 'OSC 1 / set icon name',
        '> ESC ]2<ST>': 'OSC 2 / set window title',
        '> ESC ]4<ST>': 'OSC 4 / get or set color palette',
        '> ESC ]9<ST>': 'OSC 9 / Growl integration (iTerm2)',
    }

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****


from __future__ import print_function
import sys
import os
import logging


# print version and license information
def _printver():
    import __init__

    message = '''
trachet %s
Copyright (C) 2012-2014 Hayaki Saito <user@zuse.jp>.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see http://www.gnu.org/licenses/.
    ''' % __init__.__version__

    print(message)
    return


def _parse_options():

    import optparse

    # parse options and arguments
    usage = 'usage: %prog [options] command'
    parser = optparse.OptionParser(usage=usage)

    parser.add_option('-o', '--output', dest='output',
                      default='trachetscript',
                      help='specify output device or file')

    parser.add_option('-b', '--break', dest='breakstart',
                      action="store_true", default=False,
                      help='"break" the program at the startup time'
                      '  default: false')

    parser.add_option('-m', '--monochrome', dest='monochrome',
                      action="store_true", default=False,
                      help='don\'t use color in output terminal')

    parser.add_option('-v', '--version', dest='version',
                      action="store_true", default=False,
                      help='show version')

    return parser.parse_args()


def _prepare_starting_command(args):
    # retrive starting command
    if len(args) == 1:
        command = args[0]
    elif len(args) > 1:
        command = " ".join(args)
    elif 'SHELL' in os.environ:
        command = os.getenv('SHELL')
    else:
        command = '/bin/sh'
    return command


def _prepare_term():
    # retrive TERM setting
    if 'TERM' in os.environ:
        term = os.getenv('TERM')
    else:
        term = 'xterm'
    return term


def _prepare_lang():
    # retrive LANG setting
    if 'LANG' in os.environ:
        lang = os.getenv('LANG')
    else:
        import locale
        lang = '%s.%s' % locale.getdefaultlocale()
    return lang


def _check_output_device(filename, monochrome):
    import template
    try:
        fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_NONBLOCK)
    except OSError:
        e = sys.exc_info()[1]
        logging.exception(e)
        logging.exception("Connection closed.")
        print("Cannot access output file or device.\n(%s)" % filename)
        return False

    try:
        if os.isatty(fd):
            if not os.path.exists(filename):
                print("The output device %s is not found." % filename)
                return False
            if filename == os.ttyname(0):
                print ("The output device %s is busy (current TTY). "
                       "Please specify another TTY device.") % filename
                return False
            if monochrome:
                template.disable_color()
            else:
                template.enable_color()
        else:
            template.disable_color()
    finally:
        os.close(fd)
    return True


def mainimpl(options, command, term, lang, termenc):
    from tffstub import tff
    import input
    import output
    import controller
    import trace

    if not _check_output_device(options.output, options.monochrome):
        return

    tty = tff.DefaultPTY(term, lang, command, sys.stdin)
    try:
        tty.fitsize()

        controller = controller.ActionController(tty)
        tracer = trace.TraceHandler(options.output,
                                    termenc,
                                    controller)

        if options.breakstart:
            controller.set_break()

        session = tff.Session(tty)
        session.start(stdin=sys.stdin,
                      stdout=sys.stdout,
                      termenc=termenc,
                      inputhandler=input.InputHandler(controller, tracer),
                      outputhandler=output.OutputHandler(controller, tracer))
    except IOError:
        e = sys.exc_info()[1]
        logging.exception(e)
        logging.exception("Connection closed.")
        print("Connection closed.")
    except Exception:
        e = sys.exc_info()[1]
        logging.exception(e)
        logging.exception("Aborted by exception.")
        print('trachet aborted by an uncaught exception.')
        print(' see $HOME/.trachet/log/log.txt.')
    finally:
        try:
            tty.restore_term()
        except Exception:
            e = sys.exc_info()[1]
            logging.exception(e)


def main():
    ''' entry point function for command line program '''
    options, args = _parse_options()

    if options.version:
        _printver()
        return
    command = _prepare_starting_command(args)
    term = _prepare_term()
    lang = _prepare_lang()

    # retrive terminal encoding setting
    import locale
    termenc = locale.getdefaultlocale()[1]

    if not termenc:
        termenc = "UTF-8"

    # fix for cygwin environment, such as utf_8_cjknarrow
    if termenc.lower().startswith("utf_8_"):
        termenc = "UTF-8"

    logdir = os.path.expanduser('~/.trachet/log')
    if not os.path.exists(logdir):
        os.makedirs(logdir)

    logfile = os.path.join(logdir, "log.txt")
    logging.basicConfig(filename=logfile, filemode="w")

    mainimpl(options, command, term, lang, termenc)


# main
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

import codecs
import time
import logging
import sys

from tffstub import tff
import template

try:
    from cStringIO import StringIO
except ImportError:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

# formatter
import esc
import csi
import ss2
import ss3
import char
import cstr
from iomode import IOMode


class MockController(object):
    pass


class SwitchOnOffTrait(object):

    """
    >>> controller = MockController
    >>> output_file = StringIO()
    >>> handler = TraceHandler(output_file, "utf-8", controller)
    >>> handler.is_disabled()
    False
    >>> handler.set_disabled()
    >>> handler.is_disabled()
    True
    >>> handler.set_enabled()
    >>> handler.is_disabled()
    False
    """
    _disabled = False

    def is_disabled(self):
        return self._disabled

    def set_disabled(self):
        self._disabled = True

    def set_enabled(self):
        self._disabled = False


class TraceHandler(tff.DefaultHandler, SwitchOnOffTrait):

    _io_mode = None
    _xterm_mouse_buffer = None
    _xterm_mouse_counter = 0

    def __init__(self, output_file, termenc, controller):
        if isinstance(output_file, str):
            output_file = open(output_file, "w")
        self._buffer = codecs.getwriter(termenc)(StringIO())
        self._output = output_file
        self.__bufferring = False
        self._controller = controller
        self._io_mode = IOMode()

    # Switch Input/Output prompt state

    def set_output(self):
        if self.is_disabled():
            return
        if self._io_mode.is_input():
            if self.__bufferring:
                self._buffer.write("\n")
                self.__bufferring = False
            self._io_mode.set_output()

    def set_input(self):
        if self.is_disabled():
            return
        if self._io_mode.is_output():
            if self.__bufferring:
                self._buffer.write("\n")
                self.__bufferring = False
            self._io_mode.set_input()

    # Override Interface tff.EventObserver

    def handle_esc(self, context, intermediate, final):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            seq = self._xterm_mouse_buffer
            self._xterm_mouse_buffer = None
            self.handle_invalid(context, seq)
        prompt = self._io_mode.get_prompt()
        formatted = esc.format_seq(intermediate,
                                   final,
                                   self._io_mode.is_input(),
                                   self,
                                   self._controller)
        if not formatted:
            return True
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        self._buffer.write(u"%s  %s\n" % (prompt, formatted))
        return False  # not handled

    def handle_csi(self, context, parameter, intermediate, final):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            seq = self._xterm_mouse_buffer
            self._xterm_mouse_buffer = None
            self.handle_invalid(context, seq)
        prompt = self._io_mode.get_prompt()
        formatted = csi.format_seq(parameter,
                                   intermediate,
                                   final,
                                   self._io_mode.is_input(),
                                   self,
                                   self._controller)
        if self._io_mode.is_input():
            if not parameter and not intermediate:
                if final == 0x4d:  # 'M'
                    self._xterm_mouse_counter = 3
                    self._xterm_mouse_buffer = [0x1b, 0x5b, final]
                    return False
                elif final == 0x74:  # 't'
                    self._xterm_mouse_counter = 2
                    self._xterm_mouse_buffer = [0x1b, 0x5b, final]
                    return False
                elif final == 0x54:  # 'T'
                    self._xterm_mouse_counter = 6
                    self._xterm_mouse_buffer = [0x1b, 0x5b, final]
                    return False
        if not formatted:
            return True
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        self._buffer.write(u"%s  %s\n" % (prompt, formatted))
        return False  # not handled

    def handle_ss2(self, context, final):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            seq = self._xterm_mouse_buffer
            self._xterm_mouse_buffer = None
            self.handle_invalid(context, seq)
        prompt = self._io_mode.get_prompt()
        formatted = ss2.format_seq(final,
                                   self._io_mode.is_input(),
                                   self,
                                   self._controller)
        if not formatted:
            return True
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        self._buffer.write(u"%s  %s\n" % (prompt, formatted))
        return False  # not handled

    def handle_ss3(self, context, final):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            seq = self._xterm_mouse_buffer
            self._xterm_mouse_buffer = None
            self.handle_invalid(context, seq)
        prompt = self._io_mode.get_prompt()
        formatted = ss3.format_seq(final,
                                   self._io_mode.is_input(),
                                   self,
                                   self._controller)
        if not formatted:
            return True
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        self._buffer.write(u"%s  %s\n" % (prompt, formatted))
        return False  # not handled

    def handle_control_string(self, context, prefix, value):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            seq = self._xterm_mouse_buffer
            self._xterm_mouse_buffer = None
            self.handle_invalid(context, seq)
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        prompt = self._io_mode.get_prompt()
        formatted = cstr.format_seq(prefix,
                                    value,
                                    self._io_mode.is_input(),
                                    self,
                                    self._controller)
        if not formatted:
            return True
        self._buffer.write(u"%s  %s\n" % (prompt, formatted))
        return False  # not handled

    def handle_char(self, context, c):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            buf = self._xterm_mouse_buffer
            buf.append(c)
            self._xterm_mouse_counter -= 1
            if self._xterm_mouse_counter == 0:
                self._xterm_mouse_buffer = None
                self.handle_xterm_mouse(context, buf)
            return False
        mnemonic, handled = char.format_seq(c,
                                            self._io_mode.is_input(),
                                            self,
                                            self._controller)
        if not mnemonic:
            return True
        if self.is_disabled():
            return False
        if not self.__bufferring:
            self.__bufferring = True
            prompt = self._io_mode.get_prompt()
            self._buffer.write(u"%s  " % prompt)
        if handled:
            self.__bufferring = False
            self._buffer.write(mnemonic)
            self._buffer.write("\n")
        else:
            self._buffer.write(mnemonic)
        return False  # not handled

    def handle_invalid(self, context, seq):
        if self._xterm_mouse_buffer is not None and self._io_mode.is_input():
            seq = self._xterm_mouse_buffer
            self._xterm_mouse_buffer = None
            self.handle_invalid(context, seq)
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        prompt = self._io_mode.get_prompt()
        value = str([hex(c) for c in seq])
        template_invalid = template.getinvalid()
        self._buffer.write(template_invalid % (prompt, value))

    def handle_xterm_mouse(self, context, seq):
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        prompt = self._io_mode.get_prompt()
        if seq[2] == 0x4d:  # M
            info = (seq[3] - 32, seq[4] - 32, seq[5] - 32)
            value = "xterm normal mouse: button=%d, x=%d, y=%d" % info
            params = (prompt, seq[3], seq[4], seq[5], value)
            self._buffer.write(unicode(template.getmouse()) % params)
        elif seq[2] == 0x74:  # t
            info = (seq[3] - 32, seq[4] - 32)
            value = "xterm highlight mouse: x=%d, y=%d" % info
            params = (prompt, seq[3], seq[4], value)
            self._buffer.write(unicode(template.gethighlightmouseinitial()) % params)
        elif seq[2] == 0x54:  # T
            info = (seq[3] - 32, seq[4] - 32, seq[5] - 32,
                    seq[6] - 32, seq[7] - 32, seq[8] - 32)
            value = ('xterm highlight mouse: startx=%d starty=%d, '
                     'endx=%d, endy=%d, mousex=%d, mousey=%d') % info
            params = (prompt, seq[3], seq[4], seq[5],
                      seq[6], seq[7], seq[8], value)
            self._buffer.write(unicode(template.gethighlightmouse()) % params)

    def handle_resize(self, context, row, col):
        if self.is_disabled():
            return False
        if self.__bufferring:
            self._buffer.write("\n")
            self.__bufferring = False
        prompt = "==="
        self._buffer.write(template.getresize() % (prompt, row, col))

    def handle_draw(self, context):
        try:
            self._output.write(self._buffer.getvalue())
        except IOError:
            e = sys.exc_info()[1]
            logging.exception(e)
            time.sleep(0.1)
            self._output.write(self._buffer.getvalue())
        try:
            self._output.flush()
        except IOError:
            e = sys.exc_info()[1]
            logging.exception(e)
            time.sleep(0.1)
            self._output.flush()

        self._buffer.truncate(0)


def _test():
    pass


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

from __future__ import print_function
import seqdb
import template

_DB = seqdb.get()


def get_mnemonic(direction, f):
    """
    >>> get_mnemonic('=', 'O')
    '<unknown>'
    """

    key = "%s ESC N %s" % (direction, f)
    if key in _DB:
        mnemonic = _DB[key]
    else:
        mnemonic = '<unknown>'
    return mnemonic


def format_seq(final, is_input, tracer, controller):
    f = chr(final)

    if is_input:
        direction = '<'
    else:
        direction = '>'

    mnemonic = get_mnemonic(direction, f)
    if mnemonic[0] == "!":
        return eval(mnemonic[1:])

    context = []
    if f:
        context.append(f)
    result = template.getss2() % (" ".join(context), mnemonic)
    return result


def _test():
    """
    >>> _test()
    test
    <unknown>
    """
    global _DB
    _DB = {'> ESC N O': 'test'}

    print(get_mnemonic('>', 'O'))

    print(get_mnemonic('>', 'A'))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

import seqdb
import template

_DB = seqdb.get()

_CHAR_MAP = {0x00: '<NUL>',
             0x01: '<SOH>',
             0x02: '<STX>',
             0x03: '<ETX>',
             0x04: '<EOT>',
             0x05: '<ENQ>',
             0x06: '<ACK>',
             0x07: '<BEL>',
             0x08: '<BS>',
             0x09: '<HT>',
             0x0A: '<NL>',
             0x0B: '<VT>',
             0x0C: '<NP>',
             0x0D: '<CR>',
             0x0E: '<SO>',
             0x0F: '<SI>',
             0x10: '<DLE>',
             0x11: '<DC1>',
             0x12: '<DC2>',
             0x13: '<DC3>',
             0x14: '<DC4>',
             0x15: '<NAK>',
             0x16: '<SYN>',
             0x17: '<ETB>',
             0x18: '<CAN>',
             0x19: '<EM>',
             0x1A: '<SUB>',
             0x1B: '<ESC>',
             0x1C: '<FS>',
             0x1D: '<GS>',
             0x1E: '<RS>',
             0x1F: '<US>',
             0x20: '<SP>',
             0x7F: '<DEL>'}


def get_mnemonic(key):
    """
      >>> _create_mock_db()
      >>> get_mnemonic('< <BEL>')
      'BEL / Ctrl-G'
      >>> get_mnemonic('< <DEL>')
      ''
    """
    if key in _DB:
        return _DB[key]
    return ""


def format_seq(c, is_input, tracer, controller):
    """
      >>> _create_mock_db()
      >>> template.enable_color()
      >>> str(format_seq(ord("a"), False, None, None)).replace("\x1b", "\\x1b")
      "(u'\\\\x1b[32ma\\\\x1b[m', False)"
      >>> str(format_seq(ord("\x1b"), False, None, None)).replace("\x1b", "\\x1b")
      "('\\\\x1b[32m<ESC>\\\\x1b[m', False)"
      >>> str(format_seq(ord("\x07"), False, None, None)).replace("\x1b", "\\x1b")
      "('\\\\x1b[31m<BEL>\\\\x1b[1;32m\\\\r\\\\x1b[30CBEL / bell\\\\x1b[m', True)"
    """

    if c in _CHAR_MAP:
        printable_char = _CHAR_MAP[c]
    elif c < 0x10000:
        printable_char = unichr(c)
    else:
        c -= 0x10000
        c1 = (c >> 10) + 0xd800
        c2 = (c & 0x3ff) + 0xdc00
        printable_char = unichr(c1) + unichr(c2)

    if is_input:
        direction = '<'
    else:
        direction = '>'

    key = "%s %s" % (direction, printable_char)

    mnemonic = get_mnemonic(key)
    if mnemonic:
        if mnemonic[0] == "!":
            return eval(mnemonic[1:])
        return template.getprintablechar() % (printable_char, mnemonic), True
    return template.getchar() % printable_char, False


def _create_mock_db():
    global _DB
    _DB = {
        '< <NUL>': 'NUL / Ctrl-@,Ctrl-SP,Ctrl-2',
        '< <BEL>': 'BEL / Ctrl-G',
        '> <NUL>': 'NUL / null character',
        '> <BEL>': 'BEL / bell',
    }


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****


_DEBUG_MODE_NONE = 0
_DEBUG_MODE_NORMAL_STEP = 1
_DEBUG_MODE_FUZZY_STEP = 2
_DEBUG_MODE_STOP = 3

import constant
import time


class ActionController(object):

    __mode = _DEBUG_MODE_NONE
    __actions = None
    __accel = 1.0
    __lastcalltime = 0

    def __init__(self, tty):
        self.__mode = _DEBUG_MODE_NONE
        self.__actions = []
        self.__tty = tty
        self.__accel = 1.0

    def is_suspended(self):
        return self.__mode != _DEBUG_MODE_NONE

    def append(self, action):
        return self.__actions.append(action)

    def resume(self):
        self.__mode = _DEBUG_MODE_NONE
        # self.__tty.xon()

    def set_normal_step(self):
        self.__mode = _DEBUG_MODE_NORMAL_STEP

    def set_fuzzy_step(self):
        self.__mode = _DEBUG_MODE_FUZZY_STEP

    def set_break(self):
        self.__mode = _DEBUG_MODE_STOP
        # self.__tty.xoff()

    def _get_repeat_count(self):
        now = time.time()
        if now - self.__lastcalltime < 0.1:
            self.__accel *= 1.2
        else:
            self.__accel = 1
        self.__lastcalltime = now

        repeat = max(1, self.__accel)

        return repeat

    def tick(self):
        if self.__mode == _DEBUG_MODE_NONE:
            while self.__actions:
                action = self.__actions.pop(0)
                result = action()

        elif self.__mode == _DEBUG_MODE_NORMAL_STEP:
            self.__mode = _DEBUG_MODE_STOP
            repeat = self._get_repeat_count()
            while repeat > 0 and self.__actions:
                repeat -= 1
                action = self.__actions.pop(0)
                result = action()

        elif self.__mode == _DEBUG_MODE_FUZZY_STEP:
            self.__mode = _DEBUG_MODE_STOP
            repeat = self._get_repeat_count()
            while repeat > 0:
                repeat -= 1
                while self.__actions:
                    action = self.__actions.pop(0)
                    result = action()
                    if result != constant.SEQ_TYPE_CHAR:
                        break
                else:
                    return


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

from tffstub import tff
import constant


class OutputHandler(tff.DefaultHandler):

    def __init__(self, controller, tracer):
        self._tracer = tracer
        self._controller = controller

    def handle_csi(self, context, parameter, intermediate, final):
        def action():
            context.put(0x1b)  # ESC
            context.put(0x5b)  # [
            for c in parameter:
                context.put(c)
            for c in intermediate:
                context.put(c)
            context.put(final)
            self._tracer.set_output()
            self._tracer.handle_csi(context, parameter, intermediate, final)
            return constant.SEQ_TYPE_CSI
        self._controller.append(action)
        return True  # handled

    def handle_esc(self, context, intermediate, final):
        def action():
            context.put(0x1b)  # ESC
            for c in intermediate:
                context.put(c)
            context.put(final)
            self._tracer.set_output()
            self._tracer.handle_esc(context, intermediate, final)
            return constant.SEQ_TYPE_ESC
        self._controller.append(action)
        return True  # handled

    def handle_ss2(self, context, final):
        def action():
            context.put(0x1b)  # ESC
            context.put(0x4e)  # N
            context.put(final)
            self._tracer.set_output()
            self._tracer.handle_ss2(context, final)
            return constant.SEQ_TYPE_SS2
        self._controller.append(action)
        return True  # handled

    def handle_ss3(self, context, final):
        def action():
            context.put(0x1b)  # ESC
            context.put(0x4f)  # O
            context.put(final)
            self._tracer.set_output()
            self._tracer.handle_ss3(context, final)
            return constant.SEQ_TYPE_SS3
        self._controller.append(action)
        return True  # handled

    def handle_control_string(self, context, prefix, value):
        def action():
            context.put(0x1b)  # ESC
            context.put(prefix)
            for c in value:
                context.put(c)
            context.put(0x1b)  # ESC
            context.put(0x5c)  # \
            self._tracer.set_output()
            self._tracer.handle_control_string(context, prefix, value)
            return constant.SEQ_TYPE_STR
        self._controller.append(action)
        return True

    def handle_char(self, context, final):
        def action():
            context.put(final)
            self._tracer.set_output()
            self._tracer.handle_char(context, final)
            return constant.SEQ_TYPE_CHAR
        self._controller.append(action)
        return True  # handled

    def handle_invalid(self, context, seq):
        def action():
            for c in seq:
                context.put(c)
            self._tracer.set_output()
            self._tracer.handle_invalid(context, seq)
            return constant.SEQ_TYPE_CHAR
        self._controller.append(action)
        return True  # handled

    def handle_resize(self, context, row, col):
        self._tracer.handle_resize(context, row, col)

    def handle_draw(self, context):
        self._controller.tick()
        self._tracer.handle_draw(context)


if __name__ == "__main__":
    import doctest
    doctest.testmod()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

from __future__ import print_function
import seqdb
import template

_DB = seqdb.get()


def get_mnemonic(direction, f):
    """
    >>> _create_mock_db()
    >>> get_mnemonic('<', 'A')
    'Cursor key(application keypad): up arrow'
    >>> get_mnemonic('>', 'B')
    '<unknown>'
    >>> get_mnemonic('<', '[')
    'alternate escape key'
    """
    key = "%s ESC O %s" % (direction, f)
    if key in _DB:
        mnemonic = _DB[key]
    else:
        mnemonic = '<unknown>'
    return mnemonic


def format_seq(final, is_input, tracer, controller):
    """
    >>> _create_mock_db()
    >>> template.enable_color()
    >>> format_seq(0x42, True, None, None)
    '\\x1b[0;1;36;44m ESC O B \\x1b[0;1;31m\\r\\x1b[30CCursor key(application keypad): down arrow'
    >>> template.disable_color()
    >>> format_seq(0x42, True, None, None)
    ' ESC O B   Cursor key(application keypad): down arrow'
    >>> format_seq(0x42, False, None, None)
    ' ESC O B   <unknown>'
    >>> import sys
    >>> format_seq(0x74, True, sys, None)
    test
    """
    f = chr(final)

    if is_input:
        direction = '<'
    else:
        direction = '>'

    mnemonic = get_mnemonic(direction, f)
    if mnemonic[0] == "!":
        eval(mnemonic[1:])
        return None

    context = []
    if f:
        context.append(f)
    result = template.getss3() % (" ".join(context), mnemonic)
    return result


def _test():
    """
    >>> _test()
    test
    <unknown>
    """
    global _DB
    _DB = {'> ESC O O': 'test'}

    print(get_mnemonic('>', 'O'))

    print(get_mnemonic('>', 'A'))


def _create_mock_db():
    global _DB
    _DB = {
        '> ESC O'   : 'SS3',
        '< ESC O'   : 'SS3',
        '< ESC O A' : 'Cursor key(application keypad): up arrow',
        '< ESC O B' : 'Cursor key(application keypad): down arrow',
        '< ESC O C' : 'Cursor key(application keypad): right arrow',
        '< ESC O D' : 'Cursor key(application keypad): left arrow',
        '< ESC O P' : 'F1 key (xterm)',
        '< ESC O Q' : 'F2 key (xterm)',
        '< ESC O R' : 'F3 key (xterm)',
        '< ESC O S' : 'F4 key (xterm)',
        '< ESC O t' : '!tracer.stdout.write("test")',
        '< ESC O [' : 'alternate escape key',
    }


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012-2014  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

import seqdb
import template

_DB = seqdb.get()


def get_mnemonic(direction, prefix, p, i, f):
    '''
      >>> _create_mock_db()
      >>> get_mnemonic('<', '?', '1;2', '', 'c')
      'DA1 Response'
      >>> get_mnemonic('<', '', '32;41;42', '', 'M')
      'urxvt (1015) mouse reporting (code=(32-32),row=41,col=42)'
    '''

    params = p.split(";")

    if len(p) == 0 or len(p) == len(prefix):
        length = 0
    else:
        key = '%s CSI %s%s%s' % (direction, p, i, f)
        if key in _DB:
            return _DB[key]
        length = len(params)

#    if length > 0:
#        key = '%s CSI %s;[*]%s%s' % (direction, params[0], i, f)
#        if key in _DB:
#            return _DB[key]

    key = '%s CSI %s[%s]%s%s' % (direction, prefix, length, i, f)
    if key in _DB:
        if length > 0:
            return _DB[key] % tuple(params)
        else:
            return _DB[key]

    if length > 1:
        for x in xrange(0, length):
            pbytes = ";".join(params[:x])
            key = '%s CSI %s;[%s]%s%s' % (direction, pbytes, length - x, i, f)
            if key in _DB:
                return _DB[key] % tuple(params[x:])

    key = '%s CSI %s%s%s' % (direction, prefix, i, f)
    if key in _DB:
        return _DB[key]
    return '<Unknown>'


def format_seq(parameter, intermediate, final, is_input, tracer, controller):
    '''
      >>> _create_mock_db()
      >>> template.disable_color()
      >>> format_seq([0x3f, 0x31, 0x3b, 0x32], [], 0x63, True, None, None)
      ' CSI ?1;2c    DA1 Response'
    '''

    p = ''.join([chr(c) for c in parameter])
    i = ''.join([chr(c) for c in intermediate]).replace(" ", "<SP>")
    f = chr(final)

    if is_input:
        direction = '<'
    else:
        direction = '>'

    if p and p[0] > ";":
        prefix = p[0]
    else:
        prefix = ''

    mnemonic = get_mnemonic(direction, prefix, p, i, f)
    if mnemonic[0] == "!":
        return eval(mnemonic[1:])

    return template.getcsi() % (p, i, f, mnemonic)


def _create_mock_db():
    global _DB
    _DB = {
        '< CSI A': 'Cursor key(normal keypad): up arrow',
        '< CSI 1;5A': 'Cursor key (xterm): Ctrl + up arrow',
        '< CSI [0]M': 'xterm normal mouse reporting, '
                      'following 3 bytes mean (code, row, col)',
        '< CSI [3]M': 'urxvt (1015) mouse reporting '
                      '(code=(%s-32),row=%s,col=%s)',
        '< CSI ?c': 'DA1 Response',
        '< CSI >0;95;c': 'DA2 Response: xterm patch#95 '
                         '(could be iTerm/iTerm2)',
        '> CSI @': 'ICH / insert blank characters',
        '> CSI [2]H': 'CUP / move cursor to (row=%s, col=%s)',
        '> CSI >2T': 'Title Mode - Reset (xterm) 2:'
                     ' Do not set window/icon labels using UTF-8',
    }

if __name__ == "__main__":
    import doctest
    doctest.testmod()
